# Copyright 2015 Greg M. Bernstein

"""
This is an outline of a general POX application that demonstrates:
(a) Reacting to the connection of new switches
    (i) Keep track of all the switches we have heard from
(b) Receiving packets send to the controller from a switch
    (i) Keeping track of what types of packet we have seen
(c) A method (callable from the POX command line) to create a network of dumb hubs
    (i) Installs a simple flooding flow entry to turn switches into dumb hubs
(d) A method (callable from the POX command line) to intercept particular UDP port packets
    (ii) Installs a flow into all switches to intercept UDP packets on the port you specify
(e) Listening and reacting to port status events

This application will only work with Mininet networks that are loop free since it uses flooding.

One needs to set the Python path variables so python/pox can find this module and its dependencies. From
the directory where this file exists I use the following (MS Windows):
set PYTHONPATH=.;..;
python \pox\pox.py py OutlineApp

To launch mininet using a remote controller and a topology with 3 switches and 2 hosts per switch.
sudo mn --topo linear,3,2 --controller=remote,ip=192.168.1.134

From the POX python command line we can get the instance of our class with:
app = core.OutlineApp

We can then look at the OutlineApp data (properties) or execute member functions. For example
to get some connectivity going by putting in a default "flooding" flow entry in every switch run
the make_dumb_hubs() method (from POX python command line):
app.make_dumb_hubs()

After running the above you should see "pings" in Mininet working.
With the bigger topology try the following at the mininet console:
mininet> link h1s2 s2 down
mininet> link h1s2 s2 up

You should see messages from this POX application recognizing these events.

Finally to test out UDP packet interception. From Mininet create some xterms:
mininet> xterm h1s1 h2s2 h2s3

We'll use the simple python programs in the socket_example directory for the UDP senders and receiver.
In the h2s3 xterm check the IP address. Update the IP address in the UDP send/receive scripts if needed.
Run the "python udp_receiver.py" on the h2s3 xterm.  Try running "python udp_sender.py" in the xterms for h1s1 and
h2s3.  You should see a response on the receiver xterm. If not make sure you ran make_dumb_hubs() and got the IP
addresses in the sender and receiver correct.
Then run "app.udp_intercept(5015)"  (5015 is the example UDP port make sure this port matches what was use in the
sender/receiver code). Now try running the udp_sender script. The packet shouldn't get to the receiver, but you should
see a message from the POX application that it has intercepted it.

"""

from pox.core import core
import pox.openflow.libopenflow_01 as of
import pox.lib.packet as pkt

from collections import deque  # Standard python queue structure

log = core.getLogger()


class OutlineApp(object):
    """
    Waits for OpenFlow switches to connect and then fills their forwarding tables based on
    given network topology.
    """
    def __init__(self):
        """
            Define member variables here and register the handler functions to receive open flow events.
            You do not need to keep these various up and down events.  I do this here so that we can
            look at them via the POX python command line.  I use the standard Python deque data structure
            with a finite length so that I only keep the latest few events of a given type.
        """
        self.switches = {}
        self.down_events = deque(maxlen=10)
        self.packet_in_events = deque(maxlen=10)
        self.port_events = deque(maxlen=10)
        # Let's see what kinds of packets get sent to the controller by default...
        self.packet_in_types = {pkt.ethernet.ARP_TYPE: 0, pkt.ethernet.IP_TYPE: 0, pkt.ethernet.IPV6_TYPE: 0}
        core.openflow.addListeners(self)

    def _handle_ConnectionUp(self, event):
        """ This method gets called when a switch connects to the controller.
        """
        connection = event.connection
        print "Switch {} came up".format(connection.dpid)
        print "Port information {}".format(connection.ports)
        self.switches[connection.dpid] = connection  # Save switch information

    def _handle_ConnectionDown(self, event):
        """ This is where you would handle a controller switch connectivity down event """
        print "I heard a switch {} connection go down...".format(event.dpid)
        self.down_events.appendleft(event)

    def _handle_PacketIn(self, event):
        self.packet_in_events.append(event)
        # Let's keep some counts on the types of packets received
        packet = event.parsed
        ptype = packet.type
        if ptype not in self.packet_in_types:
            self.packet_in_types[ptype] = 0
        self.packet_in_types[ptype] += 1

        # Now let's see if we intercepted a UDP packet
        if ptype != pkt.ethernet.IP_TYPE:
            return
        ip_packet = packet.payload
        print "Intercepted a packet from: {} to {}".format(ip_packet.srcip, ip_packet.dstip)
        if ip_packet.protocol != pkt.ipv4.UDP_PROTOCOL:
            return
        udp_packet = ip_packet.payload
        print "Intercepted UDP packet with source port {} and dest port {}".format(udp_packet.srcport,
                                                                                   udp_packet.dstport)

    def _handle_PortStatus(self, event):
        """Gather up some information about the event to print"""
        switch = event.dpid
        port = event.port
        print "I heard a port status change from switch {} port {}".format(switch, port)
        ofp_phy_port = event.ofp.desc
        name = ofp_phy_port.name
        if ofp_phy_port.state & 1 == 0:
            state = "Up"
        else:
            state = "Down"
        print "Port {} is {}".format(name, state)
        self.port_events.appendleft(event)

    def make_dumb_hubs(self):
        """ A command to turn wonderful OpenFlow enabled switches into dumb hubs"""
        msg = of.ofp_flow_mod()
        msg.priority = 1
        msg.actions.append(of.ofp_action_output(port=of.ofp_port_rev_map['OFPP_FLOOD']))
        for switch in self.switches.keys():  # Send the flow message to all switches
            connection = self.switches[switch]
            connection.send(msg)

    def udp_intercept(self, udp_port):
        # Set up a redirect for a particular UDP port
        # Match on Ethertype = 0x0800 IP, UDP is IP protocol number 17
        msg = of.ofp_flow_mod()
        msg.priority = 20
        udp_match = of.ofp_match(dl_type=0x800, nw_proto=17, tp_dst=udp_port)
        msg.match = udp_match
        msg.actions.append(of.ofp_action_output(port=0xfffd))  # This is the controller port
        for switch in self.switches.keys():  # Send the flow message to all switches
            connection = self.switches[switch]
            connection.send(msg)


def launch():
    """
    Starts our OutlineApp application in POX.
    """
    core.registerNew(OutlineApp)
